//============================================================================
//
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//============================================================================

module emu (
	//Master input clock
	input         CLK_50M,

	//Async reset from top-level module.
	//Can be used as initial reset.
	input         RESET,

	//Must be passed to hps_io module
	inout  [48:0] HPS_BUS,

	//Base video clock. Usually equals to CLK_SYS.
	output        CLK_VIDEO,

	//Multiple resolutions are supported using different CE_PIXEL rates.
	//Must be based on CLK_VIDEO
	output        CE_PIXEL,

	//Video aspect ratio for HDMI. Most retro systems have ratio 4:3.
	//if VIDEO_ARX[12] or VIDEO_ARY[12] is set then [11:0] contains scaled size instead of aspect ratio.
	output [12:0] VIDEO_ARX,
	output [12:0] VIDEO_ARY,

	output  [7:0] VGA_R,
	output  [7:0] VGA_G,
	output  [7:0] VGA_B,
	output        VGA_HS,
	output        VGA_VS,
	output        VGA_DE,    // = ~(VBlank | HBlank)
	output        VGA_F1,
	output [1:0]  VGA_SL,
	output        VGA_SCALER, // Force VGA scaler

	input  [11:0] HDMI_WIDTH,
	input  [11:0] HDMI_HEIGHT,
	output        HDMI_FREEZE,

`ifdef MISTER_FB
	// Use framebuffer in DDRAM (USE_FB=1 in qsf)
	// FB_FORMAT:
	//    [2:0] : 011=8bpp(palette) 100=16bpp 101=24bpp 110=32bpp
	//    [3]   : 0=16bits 565 1=16bits 1555
	//    [4]   : 0=RGB  1=BGR (for 16/24/32 modes)
	//
	// FB_STRIDE either 0 (rounded to 256 bytes) or multiple of pixel size (in bytes)
	output        FB_EN,
	output  [4:0] FB_FORMAT,
	output [11:0] FB_WIDTH,
	output [11:0] FB_HEIGHT,
	output [31:0] FB_BASE,
	output [13:0] FB_STRIDE,
	input         FB_VBL,
	input         FB_LL,
	output        FB_FORCE_BLANK,

`ifdef MISTER_FB_PALETTE
	// Palette control for 8bit modes.
	// Ignored for other video modes.
	output        FB_PAL_CLK,
	output  [7:0] FB_PAL_ADDR,
	output [23:0] FB_PAL_DOUT,
	input  [23:0] FB_PAL_DIN,
	output        FB_PAL_WR,
`endif
`endif

	output        LED_USER,  // 1 - ON, 0 - OFF.

	// b[1]: 0 - LED status is system status OR'd with b[0]
	//       1 - LED status is controled solely by b[0]
	// hint: supply 2'b00 to let the system control the LED.
	output  [1:0] LED_POWER,
	output  [1:0] LED_DISK,

	// I/O board button press simulation (active high)
	// b[1]: user button
	// b[0]: osd button
	output  [1:0] BUTTONS,

	input         CLK_AUDIO, // 24.576 MHz
	output [15:0] AUDIO_L,
	output [15:0] AUDIO_R,
	output        AUDIO_S,   // 1 - signed audio samples, 0 - unsigned
	output  [1:0] AUDIO_MIX, // 0 - no mix, 1 - 25%, 2 - 50%, 3 - 100% (mono)

	//ADC
	inout   [3:0] ADC_BUS,

	//SD-SPI
	output        SD_SCK,
	output        SD_MOSI,
	input         SD_MISO,
	output        SD_CS,
	input         SD_CD,

	//High latency DDR3 RAM interface
	//Use for non-critical time purposes
	output        DDRAM_CLK,
	input         DDRAM_BUSY,
	output  [7:0] DDRAM_BURSTCNT,
	output [28:0] DDRAM_ADDR,
	input  [63:0] DDRAM_DOUT,
	input         DDRAM_DOUT_READY,
	output        DDRAM_RD,
	output [63:0] DDRAM_DIN,
	output  [7:0] DDRAM_BE,
	output        DDRAM_WE,

	//SDRAM interface with lower latency
	output        SDRAM_CLK,
	output        SDRAM_CKE,
	output [12:0] SDRAM_A,
	output  [1:0] SDRAM_BA,
	inout  [15:0] SDRAM_DQ,
	output        SDRAM_DQML,
	output        SDRAM_DQMH,
	output        SDRAM_nCS,
	output        SDRAM_nCAS,
	output        SDRAM_nRAS,
	output        SDRAM_nWE,

`ifdef MISTER_DUAL_SDRAM
	//Secondary SDRAM
	//Set all output SDRAM_* signals to Z ASAP if SDRAM2_EN is 0
	input         SDRAM2_EN,
	output        SDRAM2_CLK,
	output [12:0] SDRAM2_A,
	output  [1:0] SDRAM2_BA,
	inout  [15:0] SDRAM2_DQ,
	output        SDRAM2_nCS,
	output        SDRAM2_nCAS,
	output        SDRAM2_nRAS,
	output        SDRAM2_nWE,
`endif

	input         UART_CTS,
	output        UART_RTS,
	input         UART_RXD,
	output        UART_TXD,
	output        UART_DTR,
	input         UART_DSR,

	// Open-drain User port.
	// 0 - D+/RX
	// 1 - D-/TX
	// 2..6 - USR2..USR6
	// Set USER_OUT to 1 to read from USER_IN.
	input   [6:0] USER_IN,
	output  [6:0] USER_OUT,

	input         OSD_STATUS,

    input         SPI_SS2,
    input         SPI_SS3,
	 output			LED
);

///////// Default values for ports not used in this core /////////

assign ADC_BUS  = 'Z;
assign USER_OUT = '1;
assign {UART_RTS, UART_DTR} = 0; //assign {UART_RTS, UART_TXD, UART_DTR} = 0;
assign {SD_SCK, SD_MOSI, SD_CS} = 'Z;
assign {SDRAM_DQ, SDRAM_A, SDRAM_BA, SDRAM_CLK, SDRAM_CKE, SDRAM_DQML, SDRAM_DQMH, SDRAM_nWE, SDRAM_nCAS, SDRAM_nRAS, SDRAM_nCS} = 'Z;
assign {DDRAM_CLK, DDRAM_BURSTCNT, DDRAM_ADDR, DDRAM_DIN, DDRAM_BE, DDRAM_RD, DDRAM_WE} = '0;  

assign VGA_SL = 0;
assign VGA_F1 = 0;
assign VGA_SCALER = 0;
assign HDMI_FREEZE = 0;

assign AUDIO_S = 0;
assign AUDIO_MIX = 0;

assign LED_DISK = 0;
assign LED_POWER = 0;
assign BUTTONS = 0;

//////////////////////////////////////////////////////////////////

wire [1:0] ar = status[9:8];

assign VIDEO_ARX = (!ar) ? 12'd4 : (ar - 1'd1);
assign VIDEO_ARY = (!ar) ? 12'd3 : 12'd0;

`include "build_id.v" 
localparam CONF_STR = {
	"Next186;;",
	"-;",
	"F0,ROM,Load BIOS;",
	"-;",
	"O89,Aspect ratio,Original,Full Screen,[ARC1],[ARC2];",
	"-;",
	"O24,CPU Speed,Maximum,/2,/3,/4,/8,/16;",
	"O56,ISA Bus Wait,1us,2us,3us,4us;",
	"O7,Fake 286,Off,On;",
	"O8,Swap Joysticks,Off,On;",
	"O9,MIDI,MPU401,COM1;",
	"T1,NMI;",
	"-;",
	"T0,Reset;",
	"R0,Reset and close OSD;",
	"V,v",`BUILD_DATE 
};

wire forced_scandoubler;
wire  [1:0] buttons;
wire [31:0] status;
wire [10:0] ps2_key;

wire        ioctl_download;
wire        ioctl_wr;
wire [24:0] ioctl_addr;
wire [15:0] ioctl_dout;
wire [15:0] ioctl_index;
wire        ioctl_wait;

wire [21:0] gamma_bus;

wire        ps2_kbd_hps_clk_in, ps2_kbd_hps_clk_out;
wire        ps2_kbd_hps_data_in, ps2_kbd_hps_data_out;
wire        ps2_mouse_hps_clk_in, ps2_mouse_hps_clk_out;
wire        ps2_mouse_hps_data_in, ps2_mouse_hps_data_out;

hps_io #(.CONF_STR(CONF_STR), .PS2DIV(2000), .WIDE(1)) hps_io
(
	.clk_sys(clk_sys),
	.HPS_BUS(HPS_BUS),
	.EXT_BUS(),
	.gamma_bus(gamma_bus),

	.forced_scandoubler(forced_scandoubler),

	.buttons(buttons),
	.status(status),
	.status_menumask({status[5]}),
	
	.ioctl_download(ioctl_download),
	.ioctl_wr(ioctl_wr),
	.ioctl_addr(ioctl_addr),
	.ioctl_dout(ioctl_dout),
	.ioctl_index(ioctl_index),
	.ioctl_wait(ioctl_wait),

	.joystick_0(joystick_0), // [7:0] on MiST, [15:0] on MiSTer
	.joystick_1(joystick_1), // [7:0] on MiST, [15:0] on MiSTer

	.joystick_l_analog_0(joystick_analog_0),
	.joystick_r_analog_0(joystick_analog_1),

	.ps2_key(ps2_key),

	.ps2_kbd_clk_out	( ps2_kbd_hps_clk_out		),  
	.ps2_kbd_data_out	( ps2_kbd_hps_data_out		), 
	.ps2_kbd_clk_in		( ps2_kbd_hps_clk_in		),   
	.ps2_kbd_data_in	( ps2_kbd_hps_data_in		),  

	//.ps2_kbd_led_status, // [2:0] fix
	//.ps2_kbd_led_use, 	 // [2:0] fix

	.ps2_mouse_clk_out	( ps2_mouse_hps_clk_out		),   
	.ps2_mouse_data_out	( ps2_mouse_hps_data_out	),  
	.ps2_mouse_clk_in	( ps2_mouse_hps_clk_in		),    
	.ps2_mouse_data_in	( ps2_mouse_hps_data_in		),   

	// TODO fix these various hps ports
	//.switches        		( switches      ),
	//.scandoubler_disable 	( scandoubler_disable ),
	//.ypbpr           		( ypbpr         ),
	//.no_csync        		( no_csync      ),

    // interface to embedded legacy sd card wrapper
	//.sd_lba          ( sd_lba        ), // ok
	//.sd_rd           ( sd_rd         ), // ok
	//.sd_wr           ( sd_wr         ), // ok
	//.sd_ack          ( sd_ack        ), // ok
	//.sd_conf         ( sd_conf       ), // not present in MiSTer sd_card
	//.sd_sdhc         ( sd_sdhc       ), // not present in MiSTer hps_io

	//.sd_dout         ( sd_dout       ), // as sd_buff_dout in MiSTer
	//.sd_dout_strobe  ( sd_dout_strobe), // not present in MiSTer hps_io
	//.sd_din          ( sd_din        ), // as sd_buff_din in MiSTer
	//.sd_din_strobe   ( sd_din_strobe ), // not present in MiSTer hps_io

	//.sd_buff_addr    ( sd_buff_addr  ), // ok
	//.sd_ack_conf     ( sd_ack_conf   ), // not present in MiSTer sd_card

	// SD config
	.img_mounted	( img_mounted	),
	.img_size		( img_size		),
	.img_readonly	( img_readonly	),

	// SD block level access
	.sd_lba			( '{sd_lba}	),	
	.sd_rd			( sd_rd		),
	.sd_wr			( sd_wr		),
	.sd_ack			( sd_ack	),

	// SD byte level access. Signals for 2-PORT altsyncram.
	.sd_buff_addr	( sd_buff_addr		),
	.sd_buff_dout	( sd_buff_dout		),
	.sd_buff_din	( '{sd_buff_din}	),	
	.sd_buff_wr		( sd_buff_wr		)
);

///////////////////////   CLOCKS   ///////////////////////////////

wire clk_sdr;	 
wire clk_mpu; 	 
wire clk_cpu, clk_dsp;
wire clk_25 = clk_sys;  // (VGA)
wire clk_27;
wire clk_14745600 = (clk_27/1875)*1024;  // 27.31 MHz

/*
	dcm dcm_system
	.inclk0(CLOCK_27),
	.c0(SDRAM_CLK),
	.c1(clk_sdr),
	.c2(clk_cpu),
	.c3(clk_dsp)
		altpll_component.clk0_divide_by = 27,
		altpll_component.clk0_multiply_by = 100,	SDRAM_CLK = 100MHz

		altpll_component.clk1_divide_by = 27,
		altpll_component.clk1_multiply_by = 100,	clk_sdr = 100MHz

		altpll_component.clk2_divide_by = 27,
		altpll_component.clk2_multiply_by = 50,    clk_cpu = 50MHz


		altpll_component.clk3_divide_by = 27,
		altpll_component.clk3_multiply_by = 50,  clk_dsp = 50MHz

*/

/*
dcm_misc dcm_misc (
	.inclk0(CLOCK_27),
	.c0(),
	.c1(CLK14745600),
	.c2(clk_mpu)
		altpll_component.clk0_divide_by = 1875,
		altpll_component.clk0_multiply_by = 784,  unused 11.2896 MHz

		altpll_component.clk1_divide_by = 1875,
		altpll_component.clk1_multiply_by = 1024,   clk147456 = 14.7456 MHz

		altpll_component.clk2_divide_by = 9,
		altpll_component.clk2_multiply_by = 1,   clk_mpu = 3 MHz

*/

/*
dcm_cpu dcm_cpu_inst (
	.inclk0(CLOCK_27),
	.c0(clk_25)

		altpll_component.clk0_divide_by = 27,
		altpll_component.clk0_multiply_by = 25,   clk_25 = 25MHz

*/

pll pll
(
	.refclk(CLK_50M),
	.rst(0),
	.outclk_0(clk_sys),      // 25 MHz
	//.outclk_1(clk_25), 	 // 25 MHz  		
	//.outclk_2(clk_14745600), // 14.7456 MHz (RS232)
	.outclk_3(clk_mpu),  	 // 3 MHz		(MPU401)

	.outclk_4(clk_27),		 // 27 MHz
	.outclk_5(clk_cpu), 	 // 50 MHz
	.outclk_6(clk_dsp),  	 // 50 MHz
	.outclk_7(clk_sdr)  	 // 100 MHz	
);

reg reset;
always @(posedge clk_cpu) reset <= RESET | status[0] | buttons[1] | !bios_loaded;

//////////////////////////////////////////////////////////////////

wire [1:0] col = status[4:3];

wire HBlank;
wire HSync;
wire VBlank;
wire VSync;
wire ce_pix;
wire [7:0] video;

wire [5:0] r,g,b;
assign VGA_R[5:0] = r; // { r,r[5:3]};
assign VGA_G[5:0] = g; // { g,g[5:3]};
assign VGA_B[5:0] = b; // { b,b[5:3]};

system next186 (
	.clk_25(clk_sys),
	.clk_sdr(clk_sdr),
	.clk_14745600(clk_14745600),
	.clk_mpu(clk_mpu),

	.clk_cpu(clk_cpu),
	.clk_en_opl2(cen_opl2),
	.clk_en_44100(cen_44100),
	.clk_dsp(clk_dsp),

	.fake286(fake286_r2),
	.cpu_speed(cpu_speed),
	.waitstates(isawait == 0 ? 8'd50 :
	            isawait == 1 ? 8'd100 :
	            isawait == 2 ? 8'd166 : 8'd200),

	.VGA_R(r),
	.VGA_G(g),
	.VGA_B(b),

	.VGA_HSYNC(HSync),
	.VGA_VSYNC(VSync),

	.hblnk(HBlank),
	.vblnk(VBlank),

	.frame_on(),

	.sdr_n_CS_WE_RAS_CAS({SDRAM_nCS, SDRAM_nWE, SDRAM_nRAS, SDRAM_nCAS}),
	.sdr_BA(SDRAM_BA),
	.sdr_ADDR(SDRAM_A),
	.sdr_DATA(SDRAM_DQ),
	.sdr_DQM({SDRAM_DQMH, SDRAM_DQML}),

	.BTN_RESET(reset),
	.BTN_NMI(NMI),

	.RS232_DCE_RXD(),
	.RS232_DCE_TXD(),
	.RS232_EXT_RXD(com1_Rx),
	.RS232_EXT_TXD(com1_Tx),
	.MPU_RX(mpu_Rx),
	.MPU_TX(mpu_Tx),

	.SD_n_CS(sd_cs),
	.SD_DI(sd_sdi),
	.SD_CK(sd_sck),
	.SD_DO(sd_sdo),

	.AUD_L(AUDIO_L),
	.AUD_R(AUDIO_R),

	.PS2_CLK1_I(ps2_kbd_hps_clk_out),
	.PS2_CLK1_O(ps2_kbd_hps_clk_in),
	.PS2_CLK2_I(ps2_mouse_hps_clk_out),
	.PS2_CLK2_O(ps2_mouse_hps_clk_in),

	.PS2_DATA1_I(ps2_kbd_hps_data_out),
	.PS2_DATA1_O(ps2_kbd_hps_data_in),
	.PS2_DATA2_I(ps2_mouse_hps_data_out),
	.PS2_DATA2_O(ps2_mouse_hps_data_in),

	.RS232_HOST_RXD(),
	.RS232_HOST_TXD(),
	.RS232_HOST_RST(),

	.GPIO_WR(joy_wr),
	.GPIO_IN(joy),

	.I2C_SCL(), //I2C_SCLK),
	.I2C_SDA(), //I2C_SDAT)

	.BIOS_ADDR(bios_addr),
	.BIOS_DIN(bios_din),
	.BIOS_WR(bios_wr),
	.BIOS_REQ(bios_req),
	.LED(LED)

	/*
	.BIOS_ADDR(ioctl_addr),
	.BIOS_DIN(ioctl_dout),
	.BIOS_WR(ioctl_wr & ioctl_download & ioctl_index==0),
	.BIOS_REQ(ioctl_wait)	
	*/
);

/*
always @(posedge clk_25) 
begin
	reg       ioctl_downlD;
	
	ioctl_downlD <= ioctl_download;
	if (ioctl_downlD & ~ioctl_download) bios_loaded <= 1;
end
*/

wire [12:0] bios_addr;
wire [15:0] bios_din;
reg         bios_loaded = 0;


wire bios_wr = 1'b1;
wire bios_req;
wire [12:0] addr;
rom #(.DW(16), .AW(13), .FN("../186Code/BIOS/Next186.hex")) BIOS
(
	.clock  	(clk_sdr 	),
	.ce     	(bios_req	), //1'b1, bios_req
	.data_out   (bios_din	),
	.a      	(addr  		),
	.out_address(bios_addr  )
);


// MiSTer VIDEO ////////////////////////////////////////////////////////////////////////////////////////////////////

assign CLK_VIDEO = clk_14745600;
assign CE_PIXEL = 1'b1;
assign VGA_DE = ~(HBlank | VBlank);
assign VGA_HS = HSync;
assign VGA_VS = VSync;

reg  [26:0] act_cnt;
always @(posedge clk_sys) act_cnt <= act_cnt + 1'd1; 
assign LED_USER    = act_cnt[26]  ? act_cnt[25:18]  > act_cnt[7:0]  : act_cnt[25:18]  <= act_cnt[7:0];

/*
next186_video next186_video
(
	.clk(clk_vid),
	.reset(reset),
	
	.pal(status[2]),
	.scandouble(forced_scandoubler),

	.ce_pix(ce_pix),

	.HBlank(HBlank),
	.HSync(HSync),
	.VBlank(VBlank),
	.VSync(VSync),

	.video(video)
);
*//////////////

// Next_186 ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

wire        btn_nmi = status[1];
wire  [2:0] speed   = status[4:2];
wire  [1:0] isawait = status[6:5];
wire        fake286 = status[7];
wire        joyswap = status[8];
wire        midi    = ~status[9];

reg   [3:0] cpu_speed;

always @(*) begin
	case (speed)
		1: cpu_speed = 1; // /2
		2: cpu_speed = 2; // /3
		3: cpu_speed = 3; // /4
		4: cpu_speed = 7; // /8
		5: cpu_speed = 15;// /16
		default: cpu_speed = 0;
	endcase
end

// core's raw video 
wire  [5:0] core_r, core_g, core_b;
wire        core_hs, core_vs;

localparam  CPU_MHZ = 50;

assign SDRAM_CKE = 1'b1;

// user io
//wire [63:0] status;
wire  [1:0] switches;



// connections between user_io (implementing the SPI communication 
// to the io controller) and the legacy 
wire [31:0] sd_lba;
wire        sd_rd;
wire        sd_wr;
wire        sd_ack;
wire        sd_sdhc = 1'b0; 
wire        sd_dout_strobe;
wire        sd_din_strobe;
wire  [8:0] sd_buff_addr;

///////////// MiSTer for reference ///////////////////////////////////////////////////////////////////////////
//wire 		  sd_rd_plus3;
//wire 		  sd_wr_plus3;
//wire [31:0] sd_lba_plus3;
//wire [7:0]  sd_buff_din_plus3;

//wire 		  sd_rd_wd;
//wire 		  sd_wr_wd;
//wire [31:0] sd_lba_wd;
//wire [7:0]  sd_buff_din_wd;

//wire        sd_rd_mmc;
//wire        sd_wr_mmc;
//wire [31:0] sd_lba_mmc;
//wire [7:0]  sd_buff_din_mmc;

//wire [31:0] sd_lba[2] = '{plus3_fdd_ready ? sd_lba_plus3 : sd_lba_wd, sd_lba_mmc};
//wire  [1:0] sd_rd = {sd_rd_mmc, plus3_fdd_ready ? sd_rd_plus3 : sd_rd_wd};
//wire  [1:0] sd_wr = {sd_wr_mmc, plus3_fdd_ready ? sd_wr_plus3 : sd_wr_wd};
//wire  [1:0] sd_ack;
//wire  [8:0] sd_buff_addr;
//wire  [7:0] sd_buff_dout;
//wire  	  sd_buff_din; // packed [7:0]  fix
//wire        sd_buff_wr;
////////////////////////////////////////////////////////////////////////////////////////

wire        img_mounted;
wire [31:0] img_size;

wire        scandoubler_disable;
wire        ypbpr;
wire        no_csync;

//assign LED_USER = ~led_out[0]; // CPU HALT

// wire the sd card to the user port
wire sd_sck;
wire sd_cs;
wire sd_sdi;
reg  sd_sdi_r;
wire sd_sdo;

always @(posedge clk_cpu) sd_sdi_r <= sd_sdi;

sd_card sd_card (
	// connection to io controller
	.clk_sys      ( clk_sys        ),
	.sd_lba       ( sd_lba         ),
	.sd_rd        ( sd_rd          ),
	.sd_wr        ( sd_wr          ),
	.sd_ack       ( sd_ack         ),
	.sdhc      	  ( sd_sdhc        ),

	.sd_buff_dout ( sd_buff_dout   ), // was sd_dout
	.sd_buff_wr   ( sd_dout_strobe ),
	.sd_buff_din  ( sd_buff_din    ), // was sd_din
	.sd_buff_addr ( sd_buff_addr   ),

	.img_mounted  ( img_mounted    ),
	.img_size     ( img_size       ),

	// connection to local CPU
	.ss   ( sd_cs          ),
	.sck  ( sd_sck         ),
	.mosi ( sd_sdi_r       ),
	.miso ( sd_sdo         )	
);

////// BIOS DOWNLOAD /////

//wire [12:0] bios_addr;
//wire [15:0] bios_din;

//wire bios_wr = 1'b1;
//wire bios_req;
/*
bios_rom bios_rom
(
  	.clk			( clk_25		 ),
	.bios_addr		( bios_addr		 ),
    .dout			( bios_din		 ),

	.ioctl_download	( ioctl_download ),
	.ioctl_wr      	( ioctl_wr       ),
	.ioctl_addr    	( ioctl_addr     ),
	.ioctl_dout    	( ioctl_dout     )
);
*/

////////////// JOYSTICKS /////////////

wire  [7:0] joystick_0;
wire  [7:0] joystick_1;
wire [15:0] joystick_analog_0;
wire [15:0] joystick_analog_1;

reg   [7:0] joy = 8'hFF;
wire        joy_wr;
reg   [7:0] joy_cnt = 8'hFF;
reg   [7:0] joy_cnt_ce_cnt;
reg         joy_cnt_ce;
wire [15:0] joy0 = joyswap ? joystick_analog_1: joystick_analog_0;
wire [15:0] joy1 = joyswap ? joystick_analog_0: joystick_analog_1;

always @(posedge clk_cpu) begin
	joy[7:4] <= joyswap ? ~{joystick_1[5:4], joystick_0[5:4]} : ~{joystick_0[5:4], joystick_1[5:4]};
	joy_cnt_ce_cnt <= joy_cnt_ce_cnt + 1'd1;
	if (joy_cnt_ce_cnt == {cpu_speed, {4{1'b1}}}) joy_cnt_ce_cnt <= 0;
	joy_cnt_ce <= joy_cnt_ce_cnt == 0;
	if (joy_wr) begin
		joy[3:0] <= 4'b1111;
		joy_cnt <= 0;
		joy_cnt_ce_cnt <= 1;
		joy_cnt_ce <= 0;
	end else if (joy_cnt != 8'hFF) begin
		if (joy_cnt == {~joy1[15], joy1[14:8]}) joy[0] <= 0;
		if (joy_cnt == {~joy1[ 7], joy1[ 6:0]}) joy[1] <= 0;
		if (joy_cnt == {~joy0[15], joy0[14:8]}) joy[2] <= 0;
		if (joy_cnt == {~joy0[ 7], joy0[ 6:0]}) joy[3] <= 0;
		if (joy_cnt_ce) joy_cnt <= joy_cnt + 1'd1;
	end else joy[3:0] <= 0;
end

////// NEXT186 ///////////////

wire [3:0] IO;
//wire [7:0] led_out;
reg        NMI;

always @(posedge clk_25) begin
	integer nmi_cnt = 0;
	reg btn_nmi_d;
	btn_nmi_d <= btn_nmi;
	if (nmi_cnt == 0) begin
		if (~btn_nmi_d & btn_nmi) nmi_cnt <= 24'hFFFFFF;
		NMI <= 0;
	end else begin
		NMI <= 1;
		nmi_cnt <= nmi_cnt - 1'd1;
	end

end

wire com1_Rx, com1_Tx;
wire mpu_Rx, mpu_Tx;

assign UART_TXD = midi ? mpu_Tx : com1_Tx;
assign com1_Rx = midi ? 1'b1 : UART_RXD;
assign mpu_Rx = midi ? UART_RXD : 1'b1;

reg         cen_opl2; // 3.58MHz
reg  [15:0] cen_opl2_cnt;
wire [15:0] cen_opl2_cnt_next = cen_opl2_cnt + 16'd358;
always @(posedge clk_cpu) begin
	cen_opl2 <= 0;
	cen_opl2_cnt <= cen_opl2_cnt_next;
	if (cen_opl2_cnt_next >= (CPU_MHZ * 100)) begin
		cen_opl2 <= 1;
		cen_opl2_cnt <= cen_opl2_cnt_next - (CPU_MHZ * 100);
	end
end

reg         cen_44100;
reg  [31:0] cen_44100_cnt;
wire [31:0] cen_44100_cnt_next = cen_44100_cnt + 16'd44100;
always @(posedge clk_cpu) begin
	cen_44100 <= 0;
	cen_44100_cnt <= cen_44100_cnt_next;
	if (cen_44100_cnt_next >= (CPU_MHZ*1000000)) begin
		cen_44100 <= 1;
		cen_44100_cnt <= cen_44100_cnt_next - (CPU_MHZ*1000000);
	end
end

reg fake286_r, fake286_r2;
always @(posedge clk_cpu) { fake286_r, fake286_r2 } <= { fake286, fake286_r };


endmodule
